from rest_framework.exceptions import PermissionDenied
from rest_framework import generics, mixins, permissions, viewsets

from huscy.attributes import models, serializer, services
from huscy.attributes.permissions import AttributeSetPermission
from huscy.subjects.models import Subject


class AttributeSchemaView(mixins.CreateModelMixin, mixins.RetrieveModelMixin,
                          generics.GenericAPIView):
    permission_classes = (permissions.DjangoModelPermissions, )
    queryset = models.AttributeSchema.objects.all()
    serializer_class = serializer.AttributeSchemaSerializer

    def get(self, request, *args, **kwargs):
        return self.retrieve(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.create(request, *args, **kwargs)

    def get_object(self):
        return services.get_attribute_schema()


class AttributeSetViewSet(mixins.RetrieveModelMixin, mixins.UpdateModelMixin,
                          viewsets.GenericViewSet):
    permission_classes = (permissions.IsAuthenticated, AttributeSetPermission)
    queryset = Subject.objects.all()
    serializer_class = serializer.AttributeSetSerializer

    def get_object(self):
        subject = super().get_object()
        attribute_set = services.get_or_create_attribute_set(subject)
        services.filter_attributes_by_category_permissions(attribute_set, self.request.user)
        return attribute_set

    def perform_update(self, serializer):
        attribute_schema = serializer.validated_data['attribute_schema']
        for node, node_description in attribute_schema.schema['properties'].items():
            if (node in serializer.validated_data['attributes'] and
                    node_description['type'] == 'object' and
                    len(serializer.validated_data['attributes'][node]) and
                    not self.request.user.has_perm(f'attributes.change_attribute_category_{node}')):
                raise PermissionDenied('You don\'t have the permission to update attribute '
                                       f'category {node}')
        super().perform_update(serializer)



from rest_framework import serializers

from huscy.attributes import models
from huscy.attributes.services import create_attribute_schema, update_attribute_set


class AttributeSchemaSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.AttributeSchema
        fields = (
            'id',
            'created_at',
            'schema',
        )

    def create(self, validated_data):
        return create_attribute_schema(**validated_data)


class AttributeSetSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.AttributeSet
        fields = (
            'attribute_schema',
            'attributes',
        )
        write_only_fields = 'attribute_schema',

    def to_representation(self, attribute_set):
        data = super().to_representation(attribute_set)
        data['attribute_schema'] = AttributeSchemaSerializer(attribute_set.attribute_schema).data
        return data

    def update(self, attribute_set, validated_data):
        return update_attribute_set(attribute_set, **validated_data)



import jsonschema
from django.db.transaction import atomic
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType

from huscy.attributes.models import AttributeSchema, AttributeSet
from huscy.pseudonyms.services import get_or_create_pseudonym


@atomic
def create_attribute_schema(schema):
    _create_attribute_category_permissions(schema)
    return AttributeSchema.objects.create(schema=schema)


def _create_attribute_category_permissions(schema):
    content_type = ContentType.objects.get_for_model(AttributeSchema)

    for name, value in schema['properties'].items():
        if value['type'] == 'object':
            Permission.objects.get_or_create(
                codename=f'change_attribute_category_{name}',
                name=f'Can change attribute category {name}',
                content_type=content_type
            )
            Permission.objects.get_or_create(
                codename=f'view_attribute_category_{name}',
                name=f'Can view attribute category {name}',
                content_type=content_type
            )


def filter_attributes_by_category_permissions(attribute_set, user):
    for node, node_description in attribute_set.attribute_schema.schema['properties'].items():
        if all([node_description['type'] == 'object',
                node in attribute_set.attributes,
                not user.has_perm(f'attributes.view_attribute_category_{node}')]):
            attribute_set.attributes.pop(node)
    return attribute_set


def get_attribute_schema(version=None):
    queryset = AttributeSchema.objects

    if version is None:
        try:
            return queryset.latest('pk')
        except AttributeSchema.DoesNotExist:
            return queryset.create()
    else:
        return queryset.get(pk=version)


def get_or_create_attribute_set(subject):
    pseudonym = get_or_create_pseudonym(subject, 'attributes.attributeset')

    attribute_set, created = AttributeSet.objects.get_or_create(pseudonym=pseudonym.code)
    return attribute_set


def update_attribute_set(attribute_set, attributes, attribute_schema=None):
    if attribute_schema:
        if attribute_schema.pk < attribute_set.attribute_schema.pk:
            raise Exception('New version for attribute schema must be greater than or equals with '
                            'current attribute schema version.')
        attribute_set.attribute_schema = AttributeSchema.objects.get(pk=attribute_schema.pk)

    _dict_merge(attribute_set.attributes, attributes)
    jsonschema.validate(attribute_set.attributes, attribute_set.attribute_schema.schema)

    attribute_set.save()
    return attribute_set


def _dict_merge(source_dict, merge_dict):
    for key, value in merge_dict.items():
        if key in source_dict and isinstance(source_dict[key], dict) and isinstance(value, dict):
            _dict_merge(source_dict[key], value)
        else:
            source_dict[key] = value



from django.contrib.auth.models import Permission
from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_403_FORBIDDEN


def test_admin_user_can_create_attribute_schema(admin_client, attribute_schema_v2):
    response = create_attribute_schema(admin_client)

    assert HTTP_201_CREATED == response.status_code


def test_admin_user_can_retrieve_attribute_schema(admin_client, attribute_schema_v4):
    response = retrieve_attribute_schema(admin_client)

    assert HTTP_200_OK == response.status_code


def test_user_with_permission_can_create_attribute_schema(client, user, attribute_schema_v2):
    user.user_permissions.add(Permission.objects.get(codename='add_attributeschema'))

    response = create_attribute_schema(client)

    assert HTTP_201_CREATED == response.status_code


def test_user_without_permission_cannot_create_attribute_schema(client, attribute_schema_v2):
    response = create_attribute_schema(client)

    assert HTTP_403_FORBIDDEN == response.status_code


def test_user_without_permission_can_retrieve_attribute_schema(client, attribute_schema_v4):
    response = retrieve_attribute_schema(client)

    assert HTTP_200_OK == response.status_code


def test_anonymous_user_cannot_create_attribute_schema(anonymous_client, attribute_schema_v2):
    response = create_attribute_schema(anonymous_client)

    assert HTTP_403_FORBIDDEN == response.status_code


def test_anonymous_user_cannot_retrieve_attribute_schema(anonymous_client, attribute_schema_v4):
    response = retrieve_attribute_schema(anonymous_client)

    assert HTTP_403_FORBIDDEN == response.status_code


def create_attribute_schema(client):
    return client.post(
        reverse('attributeschema'),
        data={'schema': {'type': 'object', 'properties': {}}},
        format='json'
    )


def retrieve_attribute_schema(client):
    return client.get(reverse('attributeschema'))



import pytest

from huscy.attributes.serializer import AttributeSchemaSerializer, AttributeSetSerializer

pytestmark = pytest.mark.django_db


def test_attributeset_serializer(attribute_set):
    expected = {
        'attribute_schema': AttributeSchemaSerializer(attribute_set.attribute_schema).data,
        'attributes': {
            'attribute1': {},
            'attribute2': 'any string',
            'attribute3': 4.5
        },
    }

    assert expected == AttributeSetSerializer(instance=attribute_set).data



from django.db import models


def get_default_attribute_schema():
    return {
        'type': 'object',
        'properties': {},
    }


class AttributeSchema(models.Model):
    schema = models.JSONField(default=get_default_attribute_schema)
    created_at = models.DateTimeField(auto_now_add=True, editable=False)


def get_latest_attribute_schema_version():
    return AttributeSchema.objects.count()


class AttributeSet(models.Model):
    pseudonym = models.CharField(max_length=128, unique=True)
    attributes = models.JSONField(default=dict)
    attribute_schema = models.ForeignKey(AttributeSchema, on_delete=models.PROTECT,
                                         default=get_latest_attribute_schema_version)



import pytest
from django.contrib.auth.models import Permission

from huscy.attributes.models import AttributeSet
from huscy.attributes.services import (
    create_attribute_schema,
    filter_attributes_by_category_permissions,
)


@pytest.fixture
def attribute_schema():
    return create_attribute_schema({
        'type': 'object',
        'properties': {
            'attribute1': {'type': 'string'},
            'category1': {
                'type': 'object',
                'properties': {
                    'attribute2': {'type': 'string'},
                }
            },
            'category2': {
                'type': 'object',
                'properties': {
                    'attribute3': {'type': 'string'},
                }
            },
        },
    })


@pytest.fixture
def attribute_set(attribute_schema):
    return AttributeSet.objects.create(
        attribute_schema=attribute_schema,
        attributes={
            'attribute1': 'foobar',
            'category1': {'attribute2': 'foo'},
            'category2': {'attribute3': 'bar'},
        }
    )


def test_without_category_permissions(user, attribute_set):
    attribute_set = filter_attributes_by_category_permissions(attribute_set, user)

    assert attribute_set.attributes == {
        'attribute1': 'foobar'
    }


def test_one_category_permission(user, attribute_set):
    view_category1_permission = Permission.objects.get(codename='view_attribute_category_category1')
    user.user_permissions.add(view_category1_permission)

    attribute_set = filter_attributes_by_category_permissions(attribute_set, user)

    assert attribute_set.attributes == {
        'attribute1': 'foobar',
        'category1': {'attribute2': 'foo'}
    }


def test_many_category_permissions(user, attribute_set):
    permissions = Permission.objects.filter(codename__contains='view_attribute_category_category')
    user.user_permissions.add(*permissions)

    attribute_set = filter_attributes_by_category_permissions(attribute_set, user)

    assert attribute_set.attributes == {
        'attribute1': 'foobar',
        'category1': {'attribute2': 'foo'},
        'category2': {'attribute3': 'bar'}
    }



import pytest

from huscy.attributes.models import AttributeSet

pytestmark = pytest.mark.django_db


def test_defaults_for_model_fields(django_db_reset_sequences, attribute_schema_v3):
    attribute_set = AttributeSet.objects.create(pseudonym='123')

    assert attribute_set.attributes == {}
    assert attribute_set.attribute_schema == attribute_schema_v3



from django.urls import path
from rest_framework.routers import DefaultRouter

from huscy.attributes import views


router = DefaultRouter()
router.register('attributesets', views.AttributeSetViewSet, basename='attributeset')

urlpatterns = [
    path('attributeschema/', views.AttributeSchemaView.as_view(), name='attributeschema'),
]

urlpatterns += router.urls
